/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
#include <glib-unix.h>
#include <gst/app/gstappsrc.h>
#include <gst/pbutils/pbutils.h>
#include <gst/gst.h>
#include <stdio.h>

/* Default values */
#define DEF_DEV_IDX	0
#define DEF_NUM_FRAMES	2000
#define DEF_MAX_BITRATE 5000
#define DEF_OUTFILE	"result.mp4"
#define DEF_LOGO_X 	0.9
#define DEF_LOGO_Y 	0

#define TYPE_MP4	1
#define TYPE_ELEM	2
#define TYPE_H264	1
#define TYPE_H265	2
#define TYPE_AV1	3

#define MAX_IN_WIDTH 1920
#define MAX_IN_HEIGHT 1080

static GMainLoop *loop = NULL;
GstElement *pipeline;
static guint signal_intr_id;
GCond cond;

typedef struct _comp_data
{
  GstElement *comp, *upload, *enc, *capsfilter, *sink, *q1, *q2, *q3, *h264_out, *qtmux;
  GstElement *src[4], *dec[4], *download[4], *h26[4], *olay, *stmp, *parsebin[4];
  GstBin *bin;
} comp_data;
typedef struct user_choice
{
  int src;
  int model;
  int sink;
} options;


static gboolean
intr_handler (gpointer user_data)
{
  printf ("Interrupt: quitting loop ...\n");
  g_main_loop_quit (loop);
  /* remove signal handler */
  signal_intr_id = 0;
  return G_SOURCE_REMOVE;
}

struct user_config
{
  char *fpath[4];
  char *lpath;
  char *outfile;
  int devidx;
  int enc_max_bit;
  float logo_x;
  float logo_y;
  int codec_t[4];
  int width_in;
  int height_in;
  int nbuffers;
};

void print_user_options (struct user_config *uconfig);
void exit_pipeline (void);
static void usage (void);
static gchar *create_symlink (const gchar *path);

static void
print_tag_foreach (const GstTagList * tags, const gchar * tag,
    gpointer user_data)
{
  GValue val = { 0, };
  gchar *str;
  int *res = user_data;

  gst_tag_list_copy_value (&val, tags, tag);

  if (G_VALUE_HOLDS_STRING (&val))
    str = g_value_dup_string (&val);
  else
    str = gst_value_serialize (&val);

  if (!strcmp ("video-codec", tag)) {
    if (strstr (str, "H.264")) {
      *res = TYPE_H264;
    } else if (strstr (str, "H.265")) {
      *res = TYPE_H265;
    } else if (strstr (str, "AV1")) {
      *res = TYPE_AV1;
    } else {
      printf ("ERROR: Codec type %s not supported\n", str);
      *res = -1;
    }
  }
  g_free (str);
  g_value_unset (&val);
}

static int
get_codec_type_real (char *path, struct user_config *uconfig)
{

  GError *err = NULL;
  GstDiscoverer *dc;
  GstDiscovererInfo *info;
  gchar *uri;
  const GstTagList *tags;
  int val = -1, w, h;
  GList *video_streams;
  dc = gst_discoverer_new (5 * GST_SECOND, &err);

  if (!dc)
    return val;
  if (err)
    return val;
  uri = gst_filename_to_uri (path, &err);
  if (err) {
    g_object_unref (dc);
    return val;
  }

  info = gst_discoverer_discover_uri (dc, uri, &err);
  if(!GST_IS_DISCOVERER_INFO (info) || err) {
    printf("Error while discovering media file : %s\n",
            err ? err->message : "");
    return -1;
  }
  video_streams = gst_discoverer_info_get_video_streams (info);
  w = gst_discoverer_video_info_get_width (video_streams->data);
  h = gst_discoverer_video_info_get_height (video_streams->data);
  gst_discoverer_stream_info_list_free (video_streams);
  if (uconfig->width_in == -1 && uconfig->height_in == -1) {
    if (w > MAX_IN_WIDTH || h > MAX_IN_HEIGHT) {
      printf ("ERROR: All input streams should be <= 1080p\n");
      return val;
    }
    uconfig->width_in = w;
    uconfig->height_in = h;
  } else if (uconfig->width_in != w || uconfig->height_in != h) {
    printf ("ERROR: All input streams should be of same resolution and should be <= 1080p\n");
    gst_discoverer_info_unref (info);
    g_object_unref (dc);
    g_free (uri);
    return val;
  }

  tags = gst_discoverer_info_get_tags (info);
  if (tags) {
    gst_tag_list_foreach (tags, print_tag_foreach, &val);
  }
  gst_discoverer_info_unref (info);
  g_object_unref (dc);
  g_free (uri);
  return val;
}

static gchar *
create_symlink (const gchar *path)
{
  gchar *full_path, *symlink_name = NULL;

  full_path = realpath (path, NULL);
  if (full_path) {
    GRand *grand = NULL;
    symlink_name = (gchar *) g_malloc0 (30);
    grand = g_rand_new ();
    if (symlink_name && grand) {
      sprintf(symlink_name, "/tmp/input_%u.h265", g_rand_int(grand));
      if (symlink (full_path, symlink_name)) {
          perror ("Error while creating symlink: ");
          g_free (symlink_name);
          symlink_name = NULL;
      }
      g_rand_free (grand);
    }
    free (full_path);
  }
  return symlink_name;
}

static void
get_codec_type (char *path, int *codec_type, struct user_config *uconfig)
{
  gchar *symlink = NULL;
  if (g_str_has_suffix (path, ".hevc")) {
   /*
    * For few hevc streams gst-discover is detecting stream type
    * as H264 and not returning proper values, but the same file
    * is working when we rename or make a symboling link to the
    * same file with .h265 extension.
    * Hence creating a symbolic link with .h265 extension in /tmp/
    * directory, reason for creating it in /tmp is that the directory
    * provided by user may not be writable if we create symbolic link
    * in the same directory.
    */
    symlink = create_symlink (path);
  }
  if (codec_type) {
    *codec_type = get_codec_type_real (symlink ? symlink : path, uconfig);
  }
  if (symlink) {
    unlink (symlink);
    g_free (symlink);
  }
}

void
print_user_options (struct user_config *uconfig)
{
  GST_INFO ("File 1 %s", uconfig->fpath[0]);
  GST_INFO ("File 2 %s", uconfig->fpath[1]);
  GST_INFO ("File 3 %s", uconfig->fpath[2]);
  GST_INFO ("File 4 %s", uconfig->fpath[3]);
  GST_INFO ("Logo file path %s", uconfig->lpath);
  GST_INFO ("Logo x relative pos is %f", uconfig->logo_x);
  GST_INFO ("Logo y relative pos is %f", uconfig->logo_y);
  GST_INFO ("Output file name is %s", uconfig->outfile);
  GST_INFO ("Max bit rate %d", uconfig->enc_max_bit);
  GST_INFO ("Device IDx value is %d", uconfig->devidx);
  GST_INFO ("width IN = %d", uconfig->width_in);
  GST_INFO ("height IN = %d", uconfig->height_in);
}

static void
fpsdisplaysink_cb (GstElement * element, gdouble fps, gdouble droprate,
    gdouble avgfps, gpointer user_data)
{
  g_print ("Average FPS %.2f\n", avgfps);
}

static void
parsebin_cb (GstElement * element, GstPad * pad, comp_data * elms)
{
  GstCaps *caps = gst_pad_query_caps (pad, NULL);
  gchar *name = gst_element_get_name (element);

  const gchar NAME_PREFIX[] = "parsebin__";

  if (g_str_has_prefix (name, NAME_PREFIX)) {
    const gchar *media_type =
        gst_structure_get_name (gst_caps_get_structure (caps, 0));

    if (g_str_has_prefix (media_type, "video/")) {
      GstPad *sinkpad;
      GstPadLinkReturn ret;
      int idx =
          g_ascii_strtoll (name + strlen (NAME_PREFIX), NULL, 10);

      GST_DEBUG ("Pad type %s for idx %d", media_type, idx);

      sinkpad = gst_element_get_static_pad (elms->h26[idx], "sink");
      ret = gst_pad_link (pad, sinkpad);
      if (ret != GST_PAD_LINK_OK) {
        printf ("ERROR: linking parsebin to h26* (%d)\n", ret);
      }
      gst_clear_object (&sinkpad);
    }
  }

  g_clear_pointer (&name, g_free);
  gst_clear_caps (&caps);
}

static int
dec_pipe (struct user_config *uconfig, comp_data * elms, int idx)
{

  gboolean bret;
  char name[50];

  sprintf (name, "%s_%d", "parsebin_", idx);
  elms->src[idx] = gst_element_factory_make ("filesrc", NULL);
  elms->parsebin[idx] = gst_element_factory_make ("parsebin", name);
  elms->download[idx] = gst_element_factory_make ("ama_download", NULL);

  if (uconfig->codec_t[idx] == TYPE_H264) {
    elms->h26[idx] = gst_element_factory_make ("h264parse", NULL);
    elms->dec[idx] = gst_element_factory_make ("ama_h264dec", NULL);
  } else if (uconfig->codec_t[idx] == TYPE_H265) {
    elms->h26[idx] = gst_element_factory_make ("h265parse", NULL);
    elms->dec[idx] = gst_element_factory_make ("ama_h265dec", NULL);
  } else if (uconfig->codec_t[idx] == TYPE_AV1) {
    elms->h26[idx] = gst_element_factory_make ("av1parse", NULL);
    elms->dec[idx] = gst_element_factory_make ("ama_av1dec", NULL);
  } else {
    printf ("ERROR: Unsupported stream format\n");
    return -1;
  }

  g_object_set (elms->src[idx], "location", uconfig->fpath[idx], "num-buffers",
      uconfig->nbuffers, NULL);
  g_object_set (elms->dec[idx], "device", uconfig->devidx, NULL);

  if (!elms->src[idx] || !elms->dec[idx] || !elms->h26[idx]
      || !elms->parsebin[idx] || !elms->download[idx]) {
    printf ("ERROR: failed to create decode %d pipeline bin...\n\n", idx);
    return -1;
  }

  gst_bin_add_many (elms->bin, elms->src[idx], elms->parsebin[idx],
      elms->h26[idx], elms->dec[idx], elms->download[idx], NULL);
  bret = gst_element_link (elms->src[idx], elms->parsebin[idx]);
  if (bret == FALSE) {
    printf ("ERROR: Failure linking %d parsebin\n", idx);
    return -1;
  }
  bret = gst_element_link_many (elms->h26[idx], elms->dec[idx],
      elms->download[idx], NULL);
  if (bret == FALSE) {
    printf ("ERROR: Failure linking %d parser\n", idx);
    return -1;
  }
  if (!g_signal_connect (elms->parsebin[idx], "pad-added",
          G_CALLBACK (parsebin_cb), elms)) {
    printf ("ERROR: Error connecting parsebin callback\n");
    return -1;
  }
  return 0;
}

static int
comp_pipe (struct user_config *uconfig, comp_data * elms)
{
  GstCaps *caps = NULL;
  gboolean bret;
  char comp_name[10] = "mix";
  GstPad *srcpad1, *sinkpad1, *srcpad2, *sinkpad2, *srcpad3, *sinkpad3;
  int width_out = uconfig->width_in * 2, height_out = uconfig->height_in * 2;
  GstElement *base_sink;

  print_user_options (uconfig);
  /* prepare the pipeline */
  elms->capsfilter = gst_element_factory_make ("capsfilter", NULL);
  elms->enc = gst_element_factory_make ("ama_h264enc", NULL);
  elms->upload = gst_element_factory_make ("ama_upload", NULL);
  elms->h264_out = gst_element_factory_make ("h264parse", NULL);
  elms->q1 = gst_element_factory_make ("queue", NULL);
  elms->q2 = gst_element_factory_make ("queue", NULL);
  elms->q3 = gst_element_factory_make ("queue", NULL);
  elms->olay = gst_element_factory_make ("gdkpixbufoverlay", NULL);
  elms->stmp = gst_element_factory_make ("timeoverlay", NULL);
  elms->comp = gst_element_factory_make ("compositor", NULL);
  elms->qtmux = gst_element_factory_make ("qtmux", NULL);
  elms->sink = gst_element_factory_make ("fpsdisplaysink", NULL);
  base_sink = gst_element_factory_make ("filesink", NULL);

  if (!elms->capsfilter || !elms->stmp || !elms->olay || !elms->enc || !elms->q1
      || !elms->q2 || !elms->q3 || !elms->comp || !elms->sink || !elms->qtmux
      || !elms->h264_out || !elms->upload || !base_sink) {
    printf ("ERROR: failed to create final base pipeline bin...\n\n");
    return -1;
  }

  if (uconfig->lpath)
    g_object_set (elms->olay, "location", uconfig->lpath, "relative-x",
        uconfig->logo_x, "relative-y", uconfig->logo_y, NULL);
  g_object_set (elms->stmp, "valignment", 1, "halignment", 1, "draw-outline", 0,
      "font-desc", "Sans, 8", "color", 0, "time-mode", 2, NULL);
  g_object_set (elms->comp, "name", comp_name, NULL);

  g_object_set (elms->upload, "device", uconfig->devidx, NULL);

  g_object_set (elms->enc,
      "bitrate", uconfig->enc_max_bit,
      "max-bitrate", uconfig->enc_max_bit,
      NULL);

  caps =
      gst_caps_new_simple ("video/x-raw", "width", G_TYPE_INT, width_out,
      "height", G_TYPE_INT, height_out, "format", G_TYPE_STRING, "I420", NULL);
  g_object_set (elms->capsfilter, "caps", caps, NULL);
  gst_caps_unref (caps);

  g_object_set (base_sink, "location", uconfig->outfile, "sync", FALSE, NULL);
  g_object_set (elms->sink, "video-sink", base_sink, "text-overlay", FALSE,
      "sync", FALSE, "fps-update-interval", 1000, "signal-fps-measurements",
      TRUE, NULL);
  //g_object_set (elms->sink, "video-sink", base_sink, "text-overlay", FALSE, "sync", FALSE, NULL);

  gst_bin_add_many (elms->bin, elms->comp, elms->capsfilter, elms->q1,
      elms->olay, elms->q2, elms->stmp, elms->q3, elms->upload, elms->enc,
      elms->h264_out, elms->qtmux, elms->sink, NULL);


  /* Sink pipe */
  bret =
      gst_element_link_many (elms->comp, elms->capsfilter, elms->q1, elms->olay,
      elms->q2, elms->stmp, elms->q3, elms->upload, elms->enc, elms->h264_out,
      elms->qtmux, elms->sink, NULL);
  if (bret == FALSE) {
    g_error ("Failure linking Enc pipeline");
    return -1;
  }

 /*** Link request pads ***/
  /* 1st pipe */
  bret = gst_element_link (elms->download[0], elms->comp);
  if (!bret)
    return -1;
  /* 2nd pipe */
  srcpad1 = gst_element_get_static_pad (elms->download[1], "src");
  if (srcpad1 == NULL) {
    printf ("ERROR: failed to get download 2 src pad\n");
    return -1;
  }
  GST_DEBUG ("Obtained static pad %s for Download 2.",
      gst_pad_get_name (srcpad1));
  sinkpad1 = gst_element_request_pad_simple (elms->comp, "sink_%u");
  if (sinkpad1 == NULL) {
    printf ("ERROR: failed to get compositor sink pad\n");
    return -1;
  }
  GST_DEBUG ("Obtained Request pad %s for compositor.",
      gst_pad_get_name (sinkpad1));
  g_object_set (sinkpad1, "xpos", uconfig->width_in, NULL);
  if (gst_pad_link (srcpad1, sinkpad1) != GST_PAD_LINK_OK) {
    printf ("ERROR: failed to link Dec 2 with Compositor pad\n");
    gst_object_unref (sinkpad1);
    gst_object_unref (srcpad1);
  }
  /* 3rd pipe */
  srcpad2 = gst_element_get_static_pad (elms->download[2], "src");
  if (srcpad2 == NULL) {
    printf ("ERROR: failed to get download 3 src pad\n");
    return -1;
  }
  GST_DEBUG ("Obtained static pad %s for Download 3.",
      gst_pad_get_name (srcpad2));
  sinkpad2 = gst_element_request_pad_simple (elms->comp, "sink_%u");
  if (sinkpad2 == NULL) {
    printf ("ERROR: failed to get compositor sink pad\n");
    return -1;
  }
  GST_DEBUG ("Obtained Request pad %s for compositor.",
      gst_pad_get_name (sinkpad2));
  g_object_set (sinkpad2, "ypos", uconfig->height_in, NULL);
  if (gst_pad_link (srcpad2, sinkpad2) != GST_PAD_LINK_OK) {
    printf ("ERROR: failed to link Dec 3 with Compositor pad\n");
    gst_object_unref (sinkpad2);
    gst_object_unref (srcpad2);
  }
  /* 4th pipe */
  srcpad3 = gst_element_get_static_pad (elms->download[3], "src");
  if (srcpad3 == NULL) {
    printf ("ERROR: failed to get download 4 src pad\n");
    return -1;
  }
  GST_DEBUG ("Obtained static pad %s for Download 4.",
      gst_pad_get_name (srcpad3));
  sinkpad3 = gst_element_request_pad_simple (elms->comp, "sink_%u");
  if (sinkpad3 == NULL) {
    printf ("ERROR: failed to get compositor sink pad\n");
    return -1;
  }
  GST_DEBUG ("Obtained Request pad %s for compositor.",
      gst_pad_get_name (sinkpad3));
  g_object_set (sinkpad3, "xpos", uconfig->width_in, "ypos", uconfig->height_in,
      NULL);
  if (gst_pad_link (srcpad3, sinkpad3) != GST_PAD_LINK_OK) {
    printf ("ERROR: failed to link Dec 4 with Compositor pad\n");
    gst_object_unref (sinkpad3);
    gst_object_unref (srcpad3);
  }

  if (!g_signal_connect (elms->sink, "fps-measurements",
          G_CALLBACK (fpsdisplaysink_cb), elms)) {
    printf ("ERROR: Error registring fpsdisplaysink callback\n");
    return -1;
  }

  if (elms->bin == NULL) {
    printf ("ERROR: bin not created...\n");
    return -1;
  }

  return 0;
}

void
exit_pipeline (void)
{
  g_main_loop_quit (loop);
  gst_element_set_state (pipeline, GST_STATE_NULL);
  gst_element_get_state (pipeline, NULL, NULL, GST_CLOCK_TIME_NONE);
  g_main_loop_unref (loop);
  gst_object_unref (pipeline);
}

static gboolean
my_bus_callback (GstBus * bus, GstMessage * message, gpointer data)
{
  GError *err;
  gchar *debug;
  struct user_config *uconfig = (struct user_config *)data;

  switch (GST_MESSAGE_TYPE (message)) {
    case GST_MESSAGE_ERROR:
      gst_message_parse_error (message, &err, &debug);
      g_error_free (err);
      g_free (debug);
      g_main_loop_quit (loop);
      break;
    case GST_MESSAGE_EOS:
      /* end-of-stream */
      g_main_loop_quit (loop);
      printf ("End Of Stream .. !!\n");
      printf ("Output is stored in the current directory (%s)\n", uconfig->outfile);
      break;
    default:
      break;
  }
  return TRUE;
};

static void usage (void)
{
  printf ("*********************************\n");
  printf ("Usage: \n");
  printf ("./ama_xcompositor [options]\n");
  printf
      ("\nMandatory Options:\n-f <1st h264 file> -f <2nd h264 file> -f <3rd h264 file> -f <4th h264 file>\n");
  printf
      ("\nOptional Options: \n-l <png logo file>\n-b <encoder max bit rate>\n-i <device index>");
  printf
      ("\n-h <horizontal alignment for logo> | Range: 0 to 1");
  printf ("\n-v <vertical alignment for logo> | Range: 0 to 1");
  printf ("\n-o <output mp4 filename>");
  printf ("\n-n <Number of input buffers>\n");
  printf
      ("\nexample:\n./ama_xcompositor -f ./video1.mp4 -f ./video2.mp4 -f ./video3.h264 -f ./video4.h265 -l ./logo.png\n");
  printf
      ("\nFor inputs to the application, each input can not be given beyond 1080p60 resolution \n");
  printf ("*********************************\n");
}

int
main (int argc, char **argv)
{
  if (argc <= 2) {
    usage ();
    exit (0);
  }
  GstStateChangeReturn sret;
  GstBus *bus;
  int opt, file_i = 0, len, i, ret, idx = 0;
  struct user_config *uconfig;
  comp_data *elms = (comp_data *) malloc (sizeof (comp_data));

  signal_intr_id = g_unix_signal_add (SIGINT, (GSourceFunc) intr_handler, NULL);
  gst_init (NULL, NULL);

  uconfig = (struct user_config *) calloc (1, sizeof (struct user_config));
  uconfig->devidx = -1;
  uconfig->enc_max_bit = -1;
  uconfig->logo_x = -1;
  uconfig->logo_y = -1;
  uconfig->width_in = -1;
  uconfig->height_in = -1;
  uconfig->nbuffers = -1;

  while ((opt = getopt (argc, argv, ":l:f:i:b:o:h:v:n:")) != -1) {
    switch (opt) {
      case 'n':
        uconfig->nbuffers = atoi (optarg);
        if (uconfig->nbuffers < 20) {
          printf ("ERROR: Number of buffers should be minimum 20\n");
          usage ();
          exit (-1);
        }
        break;
      case 'v':
        uconfig->logo_y = strtof (optarg, NULL);
        break;
      case 'h':
        uconfig->logo_x = strtof (optarg, NULL);
        break;
      case 'i':
        uconfig->devidx = atoi (optarg);
        break;
      case 'b':
        uconfig->enc_max_bit = atoi (optarg);
        break;
      case 'o':
        len = strlen (optarg);
        uconfig->outfile = (char *) malloc ((len + 1) * sizeof (char));
        strcpy (uconfig->outfile, optarg);
        break;
      case 'l':
        if (access (optarg, F_OK) != 0) {
          printf ("ERROR: %s file doesn't exist\n", optarg);
          return -1;
        }
        len = strlen (optarg);
        uconfig->lpath = (char *) malloc ((len + 1) * sizeof (char));
        strcpy (uconfig->lpath, optarg);
        break;
      case 'f':
        if (file_i > 3)         /* ignore any further files */
          break;
        if (access (optarg, F_OK) != 0) {
          printf ("ERROR: %s file doesn't exist\n", optarg);
          return -1;
        }
        len = strlen (optarg);
        uconfig->fpath[file_i] = (char *) malloc ((len + 1) * sizeof (char));
        strcpy (uconfig->fpath[file_i], optarg);
        file_i++;
        break;
      case ':':
        printf ("ERROR: option needs a value\n");
        usage ();
        exit (-1);
      case '?':
        printf ("ERROR: unknown option: %c\n", optopt);
        usage ();
        exit (-1);
    }
  }

  /* Set default options */
  if (!uconfig->outfile) {
    len = strlen (DEF_OUTFILE) + 1;
    uconfig->outfile = (char *) malloc (len * sizeof (char));
    strcpy (uconfig->outfile, DEF_OUTFILE);
  }
  if (uconfig->logo_x < 0)
    uconfig->logo_x = DEF_LOGO_X;
  if (uconfig->logo_y < 0)
    uconfig->logo_y = DEF_LOGO_Y;
  if (uconfig->nbuffers == -1)
    uconfig->nbuffers = DEF_NUM_FRAMES;
  if (uconfig->devidx == -1)
    uconfig->devidx = DEF_DEV_IDX;
  if (uconfig->enc_max_bit == -1)
    uconfig->enc_max_bit = DEF_MAX_BITRATE;
  if (file_i != 4) {
    printf ("ERROR: Please provide exactly 4 files for compositioning\n");
    usage ();
    exit (-1);
  }

  pipeline = gst_pipeline_new ("mixing");
  elms->bin = (GstBin *) gst_bin_new ("mixing_bin");

  while (idx < 4) {
    int val = -1;
    get_codec_type (uconfig->fpath[idx], &val, uconfig);
    if (val == -1)
      return -1;

    uconfig->codec_t[idx] = val;

    ret = dec_pipe (uconfig, elms, idx);
    if (ret == -1) {
      printf ("ERROR: decode bin %d not created...\n", idx);
      return ret;
    }
    idx++;
  }

  ret = comp_pipe (uconfig, elms);
  if (ret == -1) {
    printf ("ERROR: Final bin not created...\n");
    return ret;
  }
  gst_bin_add (GST_BIN (pipeline), (GstElement *) elms->bin);

  bus = gst_pipeline_get_bus (GST_PIPELINE (pipeline));
  gst_bus_add_watch (bus, my_bus_callback, uconfig);
  gst_object_unref (bus);
  /* start the pipeline */
  sret = gst_element_set_state (pipeline, GST_STATE_PLAYING);
  if (sret == GST_STATE_CHANGE_FAILURE) {
    g_error ("ERROR: Could not change state in pipeline.\n");
    gst_object_unref (pipeline);
    return -1;
  }

  /* run the pipeline */
  loop = g_main_loop_new (NULL, TRUE);
  g_main_loop_run (loop);
  exit_pipeline ();
  if (signal_intr_id > 0)
    g_source_remove (signal_intr_id);

  gst_deinit ();
  if (uconfig->lpath)
    free (uconfig->lpath);
  free (uconfig->outfile);
  for (i = 0; i < file_i; i++) {
    free (uconfig->fpath[i]);
  }
  free (uconfig);
  return 0;
}
