/*
 * Copyright (c) 2023 Xilinx
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstamaav1enc.h"

#include <gst/pbutils/pbutils.h>

GST_DEBUG_CATEGORY_STATIC (gst_ama_av1_enc_debug);
#define GST_CAT_DEFAULT gst_ama_av1_enc_debug

#define GST_AMA_AV1_ENC_CORES_DEFAULT XMA_ENC_CORES_DEFAULT

typedef enum
{
  GST_AMA_AV1_ENC_DEVICE_TYPE_1 = XMA_ENC_DEVICE_TYPE_1,
  GST_AMA_AV1_ENC_DEVICE_TYPE_2 = XMA_ENC_DEVICE_TYPE_2,
  GST_AMA_AV1_ENC_DEVICE_TYPE_DEFAULT = GST_AMA_AV1_ENC_DEVICE_TYPE_1
} GstAMAEncLatencyMode;

#define GST_TYPE_AMA_AV1_ENC_DEVICE_TYPE (gst_ama_av1_enc_device_type_get_type ())
static GType
gst_ama_av1_enc_device_type_get_type (void)
{
  static GType type = 0;

  static const GEnumValue devices[] = {
    {GST_AMA_AV1_ENC_DEVICE_TYPE_1, "AV1Enc Type 1", "1"},
    {GST_AMA_AV1_ENC_DEVICE_TYPE_2, "AV1Enc Type 2", "2"},
    {0, NULL, NULL},
  };

  if (g_once_init_enter (&type)) {
    GType tmp = g_enum_register_static ("GstAMAAV1EncDeviceType", devices);
    g_once_init_leave (&type, tmp);
  }

  return type;
}

// Introduced "type" property(inplace of "device-type") to specify AV1 Enc type which is aligned with FFMPEG CMD line option.
// "device-type" will be deprecated in subsequent releases
typedef enum
{
  GST_AMA_AV1_ENC_TYPE_1 = XMA_ENC_DEVICE_TYPE_1,
  GST_AMA_AV1_ENC_TYPE_2 = XMA_ENC_DEVICE_TYPE_2,
  GST_AMA_AV1_ENC_TYPE_DEFAULT = GST_AMA_AV1_ENC_TYPE_1
} GstAMAEnctype;

#define GST_TYPE_AMA_AV1_ENC_TYPE (gst_ama_av1_enc_type_get_type ())
static GType
gst_ama_av1_enc_type_get_type (void)
{
  static GType type = 0;

  static const GEnumValue av1enctypes[] = {
    {GST_AMA_AV1_ENC_TYPE_1, "AV1Enc Type 1", "1"},
    {GST_AMA_AV1_ENC_TYPE_2, "AV1Enc Type 2", "2"},
    {0, NULL, NULL},
  };

  if (g_once_init_enter (&type)) {
    GType tmp = g_enum_register_static ("GstAMAAV1EncType", av1enctypes);
    g_once_init_leave (&type, tmp);
  }

  return type;
}

enum
{
  PROP_0,
  PROP_DEVICE_TYPE,
  PROP_TYPE,
  PROP_CORES,
};

typedef enum
{
  GST_AMA_AV1_PROFILE_MAIN = 200,
} GstAV1Profile;

static GstStaticPadTemplate src_template = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS
    ("video/x-av1, "
        "profile= (string)main, "
        "level= (string) { 2, 2.1, 3, 3.1, 4, 4.1, 5, 5.1, 5.2, 5.3, 6, 6.1, 6.2, 6.3 };"));

G_DEFINE_TYPE (GstAMAAV1Enc, gst_ama_av1_enc, GST_TYPE_AMA_ENC);
GST_ELEMENT_REGISTER_DEFINE (ama_av1enc, "ama_av1enc", GST_RANK_NONE, GST_TYPE_AMA_AV1_ENC);

static guint32
gst_ama_av1_enc_profile_from_string (const gchar * str)
{
  if (!g_strcmp0 (str, "main")) {
    return GST_AMA_AV1_PROFILE_MAIN;
  }

  g_assert_not_reached ();
}


static void
gst_ama_av1_enc_configure_encoder (GstAMAEnc * enc, GstCaps * outcaps,
    XmaEncoderProperties * enc_props)
{
  GstAMAAV1Enc *self = GST_AMA_AV1_ENC (enc);

  GstStructure *s = gst_caps_get_structure (outcaps, 0);
  const gchar *profile_str = gst_structure_get_string (s, "profile");
  const gchar *level_str = gst_structure_get_string (s, "level");

  enc_props->hwencoder_type = XMA_AV1_ENCODER_TYPE;

  if (profile_str) {
    enc_props->profile = gst_ama_av1_enc_profile_from_string (profile_str);
  }

  if (level_str) {
    if (!g_strcmp0 (level_str, "5.3")) {
      enc_props->level = 53;
    } else {
      /* XMA codes HEVC levels as H.264 level_idc.*/
      enc_props->level = gst_codec_utils_h264_get_level_idc (level_str);
    }
  }

  enc_props->params[enc_props->param_cnt].name = XMA_ENC_PARAM_DEVICE_TYPE;
  enc_props->params[enc_props->param_cnt].type = XMA_INT32;
  enc_props->params[enc_props->param_cnt].length = sizeof (int32_t);
  enc_props->params[enc_props->param_cnt].value = &self->device_type;
  if (self->type != GST_AMA_AV1_ENC_TYPE_DEFAULT ){
    enc_props->params[enc_props->param_cnt].value = &self->type;
  }
  enc_props->param_cnt++;

  enc_props->params[enc_props->param_cnt].name = XMA_ENC_PARAM_CORES;
  enc_props->params[enc_props->param_cnt].type = XMA_INT32;
  enc_props->params[enc_props->param_cnt].length = sizeof (int32_t);
  enc_props->params[enc_props->param_cnt].value = &self->cores;
  enc_props->param_cnt++;
}

static void
gst_ama_av1_enc_init (GstAMAAV1Enc * enc)
{
}

static void
gst_ama_av1_enc_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstAMAAV1Enc *self = GST_AMA_AV1_ENC (object);

  if (prop_id >= GST_AMA_ENC_RANGED_PROPERTY_FIRST) {
    G_OBJECT_CLASS (gst_ama_av1_enc_parent_class)->set_property (object,
        prop_id, value, pspec);
    return;
  }

  switch (prop_id) {
    case PROP_DEVICE_TYPE:
      self->device_type = g_value_get_enum (value);
      break;
    case PROP_TYPE:
      self->type = g_value_get_enum (value);
      break;
    case PROP_CORES:
      self->cores = g_value_get_uint (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ama_av1_enc_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  GstAMAAV1Enc *self = GST_AMA_AV1_ENC (object);

  if (prop_id >= GST_AMA_ENC_RANGED_PROPERTY_FIRST) {
    G_OBJECT_CLASS (gst_ama_av1_enc_parent_class)->get_property (object,
        prop_id, value, pspec);
    return;
  }

  switch (prop_id) {
    case PROP_DEVICE_TYPE:
      g_value_set_enum (value, self->device_type);
      break;
    case PROP_TYPE:
      g_value_set_enum (value, self->type);
      break;
    case PROP_CORES:
      g_value_set_uint (value, self->cores);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ama_av1_enc_class_init (GstAMAAV1EncClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
  GstAMAEncClass *amaenc_class = GST_AMA_ENC_CLASS (klass);

  gobject_class->set_property = gst_ama_av1_enc_set_property;
  gobject_class->get_property = gst_ama_av1_enc_get_property;

  amaenc_class->configure_encoder = gst_ama_av1_enc_configure_encoder;

  g_object_class_install_property (gobject_class, PROP_TYPE,
      g_param_spec_enum ("type", "Av1 enc type",
          "AV1 enc type to use", GST_TYPE_AMA_AV1_ENC_TYPE,
          GST_AMA_AV1_ENC_TYPE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_DEVICE_TYPE,       
      g_param_spec_enum ("device-type", "Device type",
          "AV1 device type to use (device-type will be deprecated in subsequent releases. Use type instead)", GST_TYPE_AMA_AV1_ENC_DEVICE_TYPE,
          GST_AMA_AV1_ENC_DEVICE_TYPE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_CORES,
      g_param_spec_uint ("cores", "Cores",
          "Number of device cores to use for encoding",
          1, 2,
          GST_AMA_AV1_ENC_CORES_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  amaenc_class->qp_max = 255;
  amaenc_class->min_qp_max = XMA_ENC_AV1_MIN_QP_MAX_VALUE;
  amaenc_class->max_qp_max = XMA_ENC_AV1_MAX_QP_MAX_VALUE;
  amaenc_class->max_qp_default = XMA_ENC_AV1_MAX_QP_DEFAULT;
  amaenc_class->b_frames_max = XMA_AV1_AMD_ENC_MAX_BF_VALUE;
  gst_ama_enc_class_init_ranged_properties (amaenc_class);

  gst_element_class_add_static_pad_template (element_class, &src_template);

  gst_element_class_set_static_metadata (element_class,
      "Xilinx AMA AV1 encoder", "Encoder/Video",
      "Xilinx AV1 Encoder", "Xilinx Inc., https://www.xilinx.com");

  GST_DEBUG_CATEGORY_INIT (gst_ama_av1_enc_debug, "ama_av1enc", 0,
      "Xilinx AMA AV1 encoder");
}
