/*
 * Copyright (c) 2023 Xilinx
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstxmaframeallocator.h"

GST_DEBUG_CATEGORY_STATIC (gst_xma_frame_allocator_debug);
#define GST_CAT_DEFAULT gst_xma_frame_allocator_debug

G_DEFINE_TYPE_WITH_CODE (GstXMAFrameAllocator, gst_xma_frame_allocator,
    GST_TYPE_ALLOCATOR,
    GST_DEBUG_CATEGORY_INIT (GST_CAT_DEFAULT, "xmaframeallocator", 0, "XMA Frame allocator"));

GST_DEFINE_MINI_OBJECT_TYPE (GstXMAFrame, gst_xma_frame);

static GstMemory *
xma_frame_mem_copy (GstMemory * mem, gssize offset, gssize size)
{
  GstXMAFrameAllocator *self = GST_XMA_FRAME_ALLOCATOR (mem->allocator);
  GstXMAFrame *frame = GST_XMA_FRAME_CAST (mem);
  XmaFrame *clone;

  clone = xma_frame_clone (frame->xma->xma, frame->frame);
  if (!clone) {
    GST_ERROR_OBJECT (self, "Failed to clone XmaFrame");
    return NULL;
  }

  return gst_xma_frame_allocator_wrap_frame (self, mem->size, frame->xma, clone);
}

static void
gst_xma_frame_allocator_init (GstXMAFrameAllocator * self)
{
  GstAllocator *alloc = GST_ALLOCATOR (self);

  alloc->mem_copy = xma_frame_mem_copy;

  GST_OBJECT_FLAG_SET (self, GST_ALLOCATOR_FLAG_CUSTOM_ALLOC);
}

GstAllocator *
gst_xma_frame_allocator_new ()
{
  return g_object_new (GST_TYPE_XMA_FRAME_ALLOCATOR, NULL);
}

GstMemory *
gst_xma_frame_allocator_wrap_frame (GstXMAFrameAllocator * self, gsize size,
    GstXma * xma, XmaFrame *xmaframe)
{
  GstXMAFrame *frame;

  frame = g_new0 (GstXMAFrame, 1);

  gst_memory_init (GST_MEMORY_CAST (frame), 0, GST_ALLOCATOR (self), NULL, size,
      0, 0, size);
  frame->xma = g_object_ref (xma);
  frame->frame = xmaframe;

  GST_LOG_OBJECT (self, "Allocate memory %p", frame);

  return GST_MEMORY_CAST (frame);
}

GstMemory *
gst_xma_frame_allocator_alloc (GstXMAFrameAllocator * self, gsize size,
    GstXma * xma, XmaFrameProperties * frame_props, gboolean dummy)
{
  return gst_xma_frame_allocator_wrap_frame (self, size, xma,
      xma_frame_alloc (xma->xma, frame_props, dummy));
}

static void
gst_xma_frame_allocator_free (GstAllocator * allocator, GstMemory * mem)
{
  GstXMAFrame *frame = GST_XMA_FRAME_CAST (mem);

  GST_LOG_OBJECT (allocator, "Free memory %p", mem);

  g_clear_pointer (&frame->frame, xma_frame_free);
  g_clear_object (&frame->xma);
  g_clear_pointer (&frame, g_free);
}

static GObject *
gst_xma_frame_allocator_constructor (GType type, guint n_construct_params,
    GObjectConstructParam *construct_params)
{
  /* Allocator is a singleton. */
  static GObject *allocator = NULL;

  if (g_once_init_enter (&allocator)) {
    GObject *o;

    o = G_OBJECT_CLASS (gst_xma_frame_allocator_parent_class)->constructor (
            type, n_construct_params, construct_params);
    g_once_init_leave (&allocator, g_object_ref_sink (o));
  }

  return g_object_ref (allocator);
}

static void
gst_xma_frame_allocator_class_init (GstXMAFrameAllocatorClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstAllocatorClass *allocator_class = GST_ALLOCATOR_CLASS (klass);

  gobject_class->constructor = gst_xma_frame_allocator_constructor;
  allocator_class->alloc = NULL;
  allocator_class->free = gst_xma_frame_allocator_free;
}

gboolean
gst_is_xma_frame (GstMemory * mem)
{
  return mem && mem->allocator &&
      g_type_is_a (G_OBJECT_TYPE (mem->allocator), GST_TYPE_XMA_FRAME_ALLOCATOR);
}

void
gst_xma_frame_init (GstXMAFrame * mem, GstAllocator * allocator,
    GstXma * xma, XmaFrame * frame)
{

}
