/*
 * Copyright (c) 2023 Xilinx
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "gstxrm.h"

#include "gstxma.h"

#ifdef HAVE_XRM
#include <xrm_interface.h>
#include <xrm_enc_interface.h>
#include <xrm_dec_interface.h>
#include <xrm_scale_interface.h>

GST_DEBUG_CATEGORY_STATIC (gst_xrm_debug);
#define GST_CAT_DEFAULT gst_xrm_debug
#endif

struct _GstXrm
{
#ifdef HAVE_XRM
  union {
    XrmEncodeContext encode;
    XrmDecodeContext decode;
    XrmScaleContext scale;
  } ctx;
#endif
  void (*free) (gpointer func);
};

guint32
gst_xrm_choose_device (guint32 requested_device)
{
#ifdef HAVE_XRM
  guint32 device;

  if (requested_device != GST_XMA_DEVICE_AUTO) {
    if (g_getenv ("XRM_RESERVE_ID")) {
      GST_ERROR ("Both device is specified and XRM_RESERVE_ID is set!");
      return GST_XRM_ERROR;
    }

    return requested_device;
  }

  device = xrm_interface_get_dev_index();

  return (device < 0) ? GST_XRM_ERROR : device;
#else
  return (requested_device == GST_XMA_DEVICE_AUTO) ? 0 : requested_device;
#endif
}

void
gst_xrm_free (GstXrm *self)
{
#ifdef HAVE_XRM
  if (self->free) {
    self->free (&self->ctx);
  }
#endif
  g_free (self);
}

GstXrm *
gst_xrm_reserve_enc_resources (gint device_id, gint* slice,
    const XmaEncoderProperties * props)
{
  GstXrm *result = g_new0 (GstXrm, 1);
#ifdef HAVE_XRM
  XrmInterfaceProperties xrm_props = {0};
  gboolean is_xav1 = FALSE;
  gboolean is_ull = FALSE;
  gint num_cores = 1;
  gint res;
  guint i;

  for (i = 0; i != props->param_cnt; ++i) {
    XmaParameter *param = &props->params[i];

    if (g_str_equal (param->name, XMA_ENC_PARAM_DEVICE_TYPE)) {
      g_assert (param->type == XMA_INT32 && param->length == sizeof (int32_t));

      if (props->hwencoder_type == XMA_AV1_ENCODER_TYPE) {
        is_xav1 = (*(int32_t *)param->value == XMA_ENC_DEVICE_TYPE_1);
      }
    }

    if (g_str_equal (param->name, XMA_ENC_PARAM_CORES)) {
      g_assert (param->type == XMA_INT32 && param->length == sizeof (int32_t));

      if (props->hwencoder_type == XMA_AV1_ENCODER_TYPE) {
        num_cores = *(int32_t *)param->value;
      }
    }
  }

  xrm_props.width = props->width;
  xrm_props.height = props->height;
  xrm_props.fps_num = props->framerate.numerator;
  xrm_props.fps_den = props->framerate.denominator;
  xrm_props.is_la_enabled = props->lookahead_depth != 0;
  xrm_props.enc_cores = num_cores;

  switch (props->preset)
  {
  case XMA_ENC_PRESET_FAST:
    strncpy(xrm_props.preset, "fast", sizeof(xrm_props.preset)-1);
    break;
  case XMA_ENC_PRESET_MEDIUM:
    strncpy(xrm_props.preset, "medium", sizeof(xrm_props.preset)-1);
    break;
  case XMA_ENC_PRESET_SLOW:
    strncpy(xrm_props.preset, "slow", sizeof(xrm_props.preset)-1);
    break;

  default:
    break;
  }
  is_ull = props->lookahead_depth == 0;

  res = xrm_enc_reserve (&result->ctx.encode, device_id, *slice, is_xav1, is_ull,
      &xrm_props);

  if (res == XRM_ERROR_CONNECT_FAIL) {
    GST_WARNING ("Connection to XRM failed");
  } else if (res != XRM_SUCCESS) {
    g_clear_pointer (&result, gst_xrm_free);
  } else {
    *slice = result->ctx.encode.slice_id;
    result->free = (GDestroyNotify) xrm_enc_release;
  }
#endif
  return result;
}

GstXrm *
gst_xrm_reserve_dec_resources (gint device_id, gint width, gint height,
    gint fps_num, gint fps_den)
{
  GstXrm *result = g_new0 (GstXrm, 1);
#ifdef HAVE_XRM
  XrmInterfaceProperties xrm_props = {0};
  gint res;

  xrm_props.width = width;
  xrm_props.height = height;
  xrm_props.fps_num = fps_num;
  xrm_props.fps_den = fps_den;

  res = xrm_dec_reserve (&result->ctx.decode, device_id, &xrm_props);

  if (res == XRM_ERROR_CONNECT_FAIL) {
    GST_WARNING ("Connection to XRM failed");
  } else if (res != XRM_SUCCESS) {
    g_clear_pointer (&result, gst_xrm_free);
  } else {
    result->free = (GDestroyNotify) xrm_dec_release;
  }
#endif
  return result;
}

GstXrm *
gst_xrm_reserve_scaler_resources (gint device_id, XmaScalerProperties * props)
{
  GstXrm *result = g_new0 (GstXrm, 1);
#ifdef HAVE_XRM
  XrmInterfaceProperties xrm_props = {0};
  XrmInterfaceProperties xrm_out_props[MAX_SCALER_OUTPUTS] = {0};
  gint i;
  gint res;

  xrm_props.width = props->input.width;
  xrm_props.height = props->input.height;
  xrm_props.fps_num = props->input.framerate.numerator;
  xrm_props.fps_den = props->input.framerate.denominator;

  for (i = 0; i != props->num_outputs; ++i) {
    xrm_out_props[i].width = props->output[i].width;
    xrm_out_props[i].height = props->output[i].height;
    xrm_out_props[i].fps_num = props->output[i].framerate.numerator;
    xrm_out_props[i].fps_den = props->output[i].framerate.denominator;
  }

  res = xrm_scale_reserve (&result->ctx.scale, device_id, &xrm_props,
      xrm_out_props, props->num_outputs);

  if (res == XRM_ERROR_CONNECT_FAIL) {
    GST_WARNING ("Connection to XRM failed");
  } else if (res != XRM_SUCCESS) {
    g_clear_pointer (&result, gst_xrm_free);
  } else {
    result->free = (GDestroyNotify) xrm_scale_release;
  }
#endif
  return result;
}

void
gst_xrm_init ()
{
#ifdef HAVE_XRM
  GST_DEBUG_CATEGORY_INIT (gst_xrm_debug, "xrm", 0, "Xilinx XRM interface");
#endif
}
